use std::{collections::HashMap, io::Read, path::PathBuf, sync::Arc};

use based::{
    page,
    request::{
        api::GeneratedPager, assets::DataResponse, respond_html, respond_json, RequestContext,
        StringResponse,
    },
    ui::{
        components::{
            prelude::{InfinityScroll, Timeline, TimelineElement},
            ColoredSpinner, Search, Shell,
        },
        primitives::flex::Column,
        UIWidget,
    },
};
use chrono::NaiveDate;
use maud::{html, PreEscaped};
use rocket::{get, request::FromSegments, State};

pub mod component;
use based::ui::prelude::*;
use component::*;
use serde_json::json;

use webarc::archive::{Document, DocumentIndex};
use webarc::{
    ai::{generate_embedding, remove_data_urls, EmbedStore, SearchResult},
    archive::{extract_domains, WebsiteArchive},
    conf::get_config,
    render_page,
};

const SEARCH_BAR_STYLE: &str = "w-full px-4 mb-4 py-2 text-white bg-black border-2 border-neon-blue placeholder-neon-blue focus:ring-2 focus:ring-neon-pink focus:outline-none font-mono text-lg";

pub fn WebsiteIcon(domain: &str) -> PreEscaped<String> {
    html! {
        h2 class="text-xl font-bold mb-4 -ml-2 flex items-center w-fit" {
            img class="p-2 w-10 h-10" src=(format!("/favicon/{domain}")) {};
            a href=(format!("/d/{domain}")) { (domain) };
        };
    }
}

pub fn build_timeline(
    domains: HashMap<String, Vec<String>>,
    the_day: NaiveDate,
    vanish: bool,
) -> PreEscaped<String> {
    let mut tl = Timeline();

    let yesterday = the_day - chrono::Duration::days(1);

    let mut sorted_keys: Vec<_> = domains.keys().collect();
    sorted_keys.sort();

    for key in sorted_keys {
        let (domain, paths) = (key, domains.get(key).unwrap());
        tl = tl.add_element(TimelineElement::new(
            &the_day.to_string(),
            WebsiteIcon(domain),
            Column(
                paths
                    .into_iter()
                    .map(|path| Link(&format!("/s/{domain}/{path}"), Text(&path).large()))
                    .collect(),
            )
            .gap(ScreenValue::_2),
        ));
    }

    tl = tl.add_after(InfinityScroll(
        ColoredSpinner(Purple::_500),
        &format!("/timeline?before={}", yesterday.to_string()),
    ));

    if vanish {
        tl = tl.vanish();
    }

    tl.render_with_class("")
}

pub async fn get_domains(before: &str) -> HashMap<String, Vec<String>> {
    let today = NaiveDate::parse_from_str(before, "%Y-%m-%d").unwrap();
    DocumentIndex::get_documents_of_day(today).await
}

pub async fn get_domains_lookback(
    before: &str,
    mut days: u64,
) -> (HashMap<String, Vec<String>>, NaiveDate) {
    let mut the_day = NaiveDate::parse_from_str(before, "%Y-%m-%d").unwrap();
    while days > 0 {
        let domains = get_domains(&the_day.to_string()).await;

        if !domains.is_empty() {
            return (domains, the_day);
        }

        the_day -= chrono::Duration::days(1);
        days -= 1;
    }

    (HashMap::new(), the_day)
}

#[get("/timeline?<before>")]
pub async fn timeline_route(
    ctx: RequestContext,
    shell: &State<Shell>,
    before: Option<&str>,
) -> StringResponse {
    let today = chrono::Local::now().date_naive();

    let (domains, day) = if let Some(before) = before {
        get_domains_lookback(before, 30 * 12).await
    } else {
        get_domains_lookback(&today.to_string(), 30 * 12).await
    };

    if domains.is_empty() {
        return respond_html(Nothing().0);
    }

    if ctx.is_htmx && !ctx.htmx_redirect {
        return respond_html(build_timeline(domains, day, true).0);
    }

    let tl = Screen::medium(Padding(Nothing()).all(ScreenValue::_6))
        .on(Column(vec![
            Margin(Text("Document Timeline")._3xl().extrabold())
                .left(ScreenValue::_6)
                .render_with_class(""),
            build_timeline(domains, day, false),
        ]))
        .render_with_class("");
    page!(shell, ctx, "Timeline", tl)
}

/// Get the favicon of a domain
#[get("/favicon/<domain>")]
pub async fn favicon_route(domain: &str) -> Option<DataResponse> {
    let mut buf = Vec::new();
    std::fs::File::open(std::path::Path::new("./favicon").join(domain))
        .ok()?
        .read_to_end(&mut buf)
        .ok()?;

    // TODO : Default favicon

    Some(DataResponse::new(
        buf,
        "image/x-icon".to_string(),
        Some(60 * 60 * 24 * 5),
    ))
}

/// Websites Overview
#[get("/")]
pub async fn index(
    ctx: RequestContext,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> StringResponse {
    let websites = arc.domains();

    let content = html! {
        div class="container mx-auto p-4" {

            div class="mb-4" {
                input type="search" name="query" placeholder="Search..." class=(SEARCH_BAR_STYLE)
                hx-get=("/vector_search")
                hx-target="#website_grid" hx-push-url="true" hx-swap="outerHTML" {};
            };


        div id="website_grid" {
            h1 class="text-5xl font-bold text-center mb-10" { "Websites" };
            div class="grid grid-cols-2 sm:grid-cols-3 lg:grid-cols-5 xl:grid-cols-6 2xl:grid-cols-8 gap-6" {
                @for site in websites {
                    a href=(format!("/d/{site}")) class="bg-neutral-900 shadow-md rounded-lg hover:bg-neutral-800 bg-gray-1 hover:cursor-pointer transition-all duration-300 flex flex-col items-center justify-center aspect-square max-w-60" {
                        div class="bg-blue-500 text-white rounded-full p-4" {
                            (favicon(&site))
                        };
                        p class="mt-4 text-base font-medium" { (site) };
                    };
                };
            };
        };
        }
    };

    render_page(content, ctx, &shell).await
}

/// Overview on <domain> / <path>
#[get("/d/<domain>/<paths..>")]
pub async fn domain_info_route(
    ctx: RequestContext,
    domain: &str,
    paths: PathBuf,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> StringResponse {
    let domain = arc.get_domain(domain);
    let document = domain.path(paths.to_str().unwrap());
    let versions: Vec<String> = document.versions();
    let (path_entries, is_doc) = domain.paths(paths.to_str().unwrap());
    let path_seperations: Vec<&str> = paths.to_str().unwrap().split('/').collect();

    let domains = extract_domains(
        &document
            .render_local(None, &shell)
            .await
            .unwrap_or_default(),
    );

    let content = html! {
        h2 class="text-xl font-bold mb-4 flex items-center w-fit mx-auto" {
            img class="p-2" src=(format!("/favicon/{}", &domain.name)) {};
            a href=(format!("/d/{}", &domain.name)) { (domain.name) };
            (slash_seperator())
            (gen_path_header(path_seperations, &domain.name, true))
        };

        @if !versions.is_empty() {
            div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md mb-8" {
                h3 class="font-bold mb-2" { "Saved Versions:" };
                @for (index, ver) in versions.into_iter().enumerate() {
                    a href=(format!("/s/{}/{}?time={}", &domain.name, paths.to_str().unwrap(), ver)) class="text-xs text-blue-400" {

                        @if index == 0 {
                        (format!("Latest ({ver})"))
                        } @else {
                            (ver)
                        }
                    };
                }
            };
        }

        @if path_entries.is_empty() && !is_doc {
            div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md" {
                p class="p-4 text-xl" { "Nothing found" };
            }
        }

        @if !path_entries.is_empty() {
        div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md" {
            h3 class="font-bold mb-2" { "Paths:" };
            ul class="space-y-2" {
                @for path in path_entries {
                    a href=(path.url()) class="flex items-center gap-2 p-3 border bg-neutral-800 rounded hover:shadow-lg transition" {
                        (arrow_icon("blue"))

                        span class="font-medium" { (path.path()) };
                    };
                };
            };
        };
    };

    @if !domains.is_empty() {
        div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md" {
            h3 class="font-bold mb-2" { "Domains linked on this page:" };
            ul class="space-y-2 p-4" {
                @for domain in domains {
                    a href=(format!("/d/{domain}")) class="flex items-center gap-2 p-3 border bg-neutral-800 rounded hover:shadow-lg transition" {
                        (favicon(&domain));
                        span class="font-medium" { (domain) };
                    };
                };
            };
        };
    };
    };

    render_page(content, ctx, &shell).await
}

#[get("/txt/<domain>/<path..>?<time>&<no_data_urls>")]
pub async fn render_txt_website(
    domain: &str,
    path: PathBuf,
    time: Option<&str>,
    no_data_urls: Option<&str>,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> Option<String> {
    let document = arc.get_domain(domain).path(path.to_str().unwrap());

    let mut content = document
        .render_local(time.map(|time| time.to_string()), &shell)
        .await?;

    if no_data_urls.is_some() {
        content = remove_data_urls(&content);
    }

    Some(html2md::parse_html(&content))
}

pub struct PathSegment {
    segments: Vec<String>,
}

impl std::fmt::Display for PathSegment {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(&self.to_str())
    }
}

impl PathSegment {
    pub fn to_str(&self) -> String {
        self.segments.join("/")
    }
}

impl<'r> FromSegments<'r> for PathSegment {
    type Error = ();

    fn from_segments(
        segments: rocket::http::uri::Segments<'r, rocket::http::uri::fmt::Path>,
    ) -> Result<Self, Self::Error> {
        let paths: Vec<_> = segments
            .filter_map(|x| {
                if x == "." {
                    return None;
                }

                if x == ".." {
                    return None;
                }

                Some(x.to_string())
            })
            .collect();

        Ok(PathSegment { segments: paths })
    }
}

fn is_older_than(date_str: &str, num_days: usize) -> bool {
    let date =
        chrono::NaiveDate::parse_from_str(date_str, "%Y-%m-%d").expect("Invalid date format");
    let threshold_date = chrono::Utc::now().date_naive();
    let diff = threshold_date - date;
    diff.num_days() > num_days as i64
}

pub async fn redownload(
    arc: &WebsiteArchive,
    domain: &str,
    path: PathSegment,
    document: &Document,
    shell: &Shell,
) -> Option<DataResponse> {
    arc.archive_url(&format!("https://{domain}/{}", path.to_str()))
        .await;

    let content = document.render_local(None, &shell).await?;

    return Some(DataResponse::new(
        content.as_bytes().to_vec(),
        "text/html".to_string(),
        Some(60 * 60 * 24),
    ));
}

/// Return archived version of `domain` / `path` at `time`
#[get("/s/<domain>/<path..>?<time>")]
pub async fn render_website(
    domain: &str,
    path: PathSegment,
    time: Option<&str>,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> Option<DataResponse> {
    let document = arc.get_domain(domain).path(&path.to_str());

    if time.is_none() {
        let versions = document.versions();
        let latest_version = versions.first()?;
        if let Some(outdated) = get_config().get_outdated(domain) {
            if is_older_than(latest_version, outdated) {
                log::info!("Document {domain} / {path} is outdated, redownloading");
                return redownload(&arc, domain, path, &document, &shell).await;
            }
        }
    }

    let content = document
        .render_local(time.map(|time| time.to_string()), &shell)
        .await;

    // TODO : keep n versions

    if let Some(content) = content {
        return Some(DataResponse::new(
            content.as_bytes().to_vec(),
            "text/html".to_string(),
            Some(60 * 60 * 24),
        ));
    } else if get_config().DOWNLOAD_ON_DEMAND && time.is_none() {
        return redownload(&arc, domain, path, &document, &shell).await;
    }

    None
}

pub fn gen_search_element(x: &SearchResult) -> PreEscaped<String> {
    html! {
        div class="text-xl font-bold mt-4 p-4 flex items-center w-full max-w-4xl max-h-40 mx-auto bg-neutral-800 shadow-md rounded-lg overflow-hidden border border-neutral-900 hover:cursor-pointer"
        hx-get=(format!("/d/{}/{}", x.domain, x.path))
        hx-target="#main_content" hx-push-url="true" hx-swap="innerHTML"
        {
                (favicon(&x.domain))
                a { (x.domain) };
                (slash_seperator());
                (gen_path_header(x.path.split('/').collect(), &x.domain, false));
                p class="font-bold p-2 text-stone-400" { (format!("{:.2} % [{} matches]", x.similarity() * 100.0, x.chunks.len())) };
        };
    }
}

#[get("/vector_search?<query>&<page>")]
pub async fn vector_search(
    query: Option<&str>,
    page: Option<i64>,
    ctx: RequestContext,
    shell: &State<Shell>,
) -> Option<StringResponse> {
    get_config().ai.as_ref()?;

    let page = page.unwrap_or(1);

    // Search
    let search =
        Search::new("/vector_search".to_string()).search_class(SEARCH_BAR_STYLE.to_string());

    if let Some(query) = query {
        // If we have query
        let real_query = query.trim_end_matches(".json");

        // Search Results
        let vector = Arc::new(pgvector::Vector::from(
            generate_embedding(real_query.to_string()).await?,
        ));

        let results = GeneratedPager::new(
            |input: Arc<pgvector::Vector>, offset, limit| {
                Box::pin(async move {
                    EmbedStore::search_vector(&input, limit as i64, offset as i64).await
                })
            },
            1500,
        )
        .pager(page as u64, vector)
        .await;

        // API Route
        if query.ends_with(".json") {
            // TODO : Better search API
            return Some(respond_json(&json!(&results.page(page as u64))));
        }

        let content = search.build_response(&ctx, results, page, real_query, gen_search_element);

        return Some(render_page(content, ctx, &shell).await);
    }

    // Return new search site
    let content = search.build("", html! {});
    Some(render_page(content, ctx, &shell).await)
}
